import { Subscription, PartialObserver } from '../types';
import { Data } from './internal';
export declare class Cell<T> {
    protected internal: Data<T>;
    static of<T>(initial: T): Cell<T>;
    constructor(internal: Data<T>);
    get $(): T;
    get current(): T;
    [Symbol.toStringTag]: string;
    /**
     * @deprecated Should be only used for interop.
     */
    subscribe(observer?: PartialObserver<T>): Subscription;
    /**
     * @deprecated Should be only used for interop.
     */
    subscribe(next?: (value: T) => void, error?: (error: any) => void, complete?: () => void): Subscription;
}
export declare class Reaction {
    private internal;
    constructor(callback: () => void);
    /**
     *
     * @param program
     * @returns return value of `program`
     * @throws User-land errors in program()
     */
    track<T>(program: () => T): T;
    dispose(): void;
}
export declare class IncrementalReaction {
    private internal;
    constructor(callback: () => void);
    track<T>(program: () => T): T;
    continue<T>(program: () => T): T;
    private getHistoryReads;
    dispose(): void;
}
export declare function mutable<T>(initialValue: T): [
    Cell<T>,
    (value: (T extends Function ? never : T) | ((current: T) => T)) => void
];
export declare function computed<T>(expr: () => T): Cell<T>;
declare type Unwrap<T> = {
    [Key in keyof T]: T[Key] extends Cell<infer CellType> ? CellType : T[Key];
};
export declare function combined<TArray extends any[]>(array: TArray): Cell<Unwrap<TArray>>;
export declare function combined<TObject extends object>(object: TObject): Cell<Unwrap<TObject>>;
export {};
