declare const enum BitFlags {
    /**
     * this is a data */
    Data = 1,
    /**
     * this is a memo */
    Memo = 2,
    /**
     *
     */
    Dirty = 4,
    /**
     * current value is changed (so need to propagate)
     */
    Changed = 8,
    Estimating = 16,
    /**
     * (computation) current value maybe dirty
     */
    MarkForCheck = 32,
    /**
     *
     */
    EffectWillCommit = 64,
    Effect = 256,
    /** only memo can be stale */
    StaleMemo = 1024,
    HasError = 2048
}
interface Data<T = unknown> {
    flags: BitFlags;
    lo: ObserverLinked<Memo> | null;
    value: T | null;
}
interface Memo<T = unknown> extends Data<T> {
    dc: number;
    fs: SourceLinked<Data> | null;
    ls: SourceLinked<Data> | null;
    c: () => T;
}
interface Reaction extends Memo<void> {
}
declare type SourceLinked<T> = {
    prev: SourceLinked<T> | null;
    next: SourceLinked<T> | null;
    source: T;
    observer_ref: ObserverLinked<any>;
};
declare type ObserverLinked<T> = {
    prev: ObserverLinked<T> | null;
    next: ObserverLinked<T> | null;
    observer: T;
};
declare function setData<T>(data: Data<T>, value: T): void;
/**
 * Get the value of a node _in many contexts_. Full of dirty side-effects.
 * @param data
 * @returns current value, or previous value if circularly referenced
 * @throws User-land errors, DEFER_COMPUTATION
 */
declare function accessValue<T>(data: Data<T>): T;
export declare function accessRefValue<T>(data: Data<T>): T;
declare function untrack<T, Args extends any[]>(fn: (...args: Args) => T, ...args: Args): T;
declare function cleanupMemo(memo: Memo): void;
declare class Transaction {
    dirty: Data[];
    effects: Reaction[];
    constructor();
    start(fn: () => any): any;
    flushing: boolean;
    flush(): void;
    commit(): void;
    addDirty(data: Data): void;
}
declare function batch(fn: Function): any;
declare function createReaction(onCommit: () => void): Reaction;
declare function executeReaction<T>(reaction: Reaction, fn: () => T): T;
declare function createData<T>(value: T): Data<T>;
declare function createMemo<T>(fn: () => T): Memo<T>;
export declare function __current_transaction(): Transaction | null;
export { BitFlags, Data, Memo, Reaction, batch, setData, accessValue, createData, createMemo, untrack, createReaction, executeReaction, cleanupMemo, };
