import { Subscribable, Cleanable } from '../types';
import type { Runnable, TaskResult, RunnableGenerator, TaskCompleteResult, TaskErrorResult, TaskFulfillResult } from './types';
export declare class AbortedError extends Error {
    name: string;
    constructor(message?: string);
}
export declare class TaskKilledError extends Error {
    name: string;
    constructor();
}
export declare class AbortablePromise<T> extends Promise<T> implements Runnable<T> {
    readonly abort: () => void;
    constructor(executor: (resolve: (value: T | PromiseLike<T>) => void, reject: (reason?: any) => void) => Cleanable);
    [Symbol.iterator](): RunnableGenerator<T>;
}
export declare class TaskSuspended {
    private readonly onAbort;
    private aborted;
    constructor(onAbort: () => void);
    abort(): void;
}
export declare function executeRunnableBlock<T>(block: Runnable<T>, asyncContinuation?: (result: TaskResult) => void): TaskResult;
export declare function executeRunnableTask<T>(runnable: Runnable<T>, asyncContinuation?: (result: TaskFulfillResult | TaskErrorResult) => void): TaskFulfillResult | TaskErrorResult;
export interface Task<T> extends AbortablePromise<T> {
}
declare type GeneratorReturnType<T> = T extends () => Runnable<infer R> ? R : undefined;
export declare function task<TaskFn extends (...args: any[]) => Runnable<any>>(taskFn: TaskFn): (...params: Parameters<TaskFn>) => Task<GeneratorReturnType<TaskFn>>;
export declare function delay(ms?: number): Runnable<void>;
export declare function timeout(ms?: number): Runnable<void>;
export declare function nextAnimationFrame(): Runnable<number>;
export declare function resolve<T>(value: T | Runnable<T> | PromiseLike<T> | Subscribable<T>): Runnable<T>;
export declare function __error(error: any): TaskErrorResult;
export declare function __fulfill(value: any): TaskFulfillResult;
export declare function __return(value: any): TaskCompleteResult;
export declare function __break(label?: string): TaskCompleteResult;
export declare function __continue(label?: string): TaskCompleteResult;
export declare function __handle_task_complete(result: TaskResult): TaskFulfillResult | TaskErrorResult;
export declare function __abort_all(disposers: TaskSuspended[]): void;
export {};
